#!/usr/bin/env python

import sys
import json
import emod_api.interventions.common as comm
#import emodpy_malaria.interventions.common as malcomm
import emodpy_hiv.interventions.peek_camp as pc
from emodpy_hiv.interventions.peek_camp import encode
import emod_api.campaign as camp

from emodpy_hiv.interventions.art import new_intervention as art
from emodpy_hiv.interventions.random import new_diagnostic as hrc
from emodpy_hiv.interventions.modcoinf import new_intervention as coinf
from emodpy_hiv.interventions.stipostdebut import new_diagnostic as post
from emodpy_hiv.interventions.rapiddiag import new_diagnostic as rd
from emodpy_hiv.interventions.yearandsexdiag import new_diagnostic as yasd
from emod_api.interventions.migration import _new_migration_intervention
import emodpy_hiv.interventions.outbreak as ob

def iv_class_factory_ish( iv_name, payload ):
    """
    Here's the messy part. Lots of one-off handling.
    """
    iv = None
    try:
        method_to_call = getattr(comm, iv_name)
    except Exception as ex:
        method_to_call = None
    if iv_name == "PropertyValueChanger":
        ips = payload.split( ":" )
        iv = method_to_call( camp, Target_Property_Key=ips[0], Target_Property_Value=ips[1] )
    elif iv_name == "AntiretroviralTherapy":
        iv = art( camp )
    elif iv_name == "BroadcastEvent":
        iv = method_to_call( camp, Event_Trigger=payload )
    elif iv_name == "HIVRandomChoice":
        payload = json.loads( payload.replace( "'", '"' ) )
        iv = hrc( camp, choices=payload )
    elif iv_name == "ModifyStiCoInfectionStatus":
        iv = coinf( camp )
    elif iv_name == "STIIsPostDebut":
        if not payload:
            raise ValueError( "STIIsPostDebut needs a parameter!" )
        bcast_signals = payload.split("/")
        if len(bcast_signals) == 1:
            neg = "NoTrigger"
        else:
            neg = Negative_Event=bcast_signals[1]
        iv = post( camp, Positive_Event=bcast_signals[0], Negative_Event=neg )
    elif iv_name == "HIVRapidHIVDiagnostic":
        bcast_signals = payload.split("/")
        iv = rd( camp, Positive_Event=bcast_signals[0], Negative_Event=bcast_signals[1] )
    elif iv_name == "HIVPiecewiseByYearAndSexDiagnostic":
        payload = payload.split("/")
        tvmap = pc.presets[payload[0]]
        iv = yasd( camp, TVMap=tvmap, Positive_Event=payload[1], Negative_Event=payload[2] )
    elif iv_name.startswith( "Delay" ):
        raise ValueError( "This pathway should not be invoked anymore but not 100% certain." )
    elif iv_name.startswith( "Outbreak" ):
        iv = ob._Outbreak( camp )
    elif iv_name == "MigrateIndividuals":
        iv = _new_migration_intervention( camp, nodeid_to_migrate_to=int(payload) ) # defaults won't work
        print( "NOTE: MigrateIndividuals will need more params for durations!" )
    elif iv_name == "SimpleHealthSeekingBehavior":
        payload = payload.split("/")
        event = payload[0]
        tendency = float(payload[1]) if len(payload)>1 else 1.0
        name = payload[2] if len(payload)>2 else None
        # (Event/Tendency/Name)
        iv = comm.HSB( camp, Event=event, Tendency=tendency, Name=name )
    elif iv_name == "ARTBasic":
        from emodpy_hiv.interventions.art import new_intervention as ni
        iv = ni( camp )
    elif iv_name == "ARTDropout":
        from emodpy_hiv.interventions.artdropout import new_intervention as ni
        iv = ni( camp )
    elif iv_name == "PMTCT":
        from emodpy_hiv.interventions.pmtct import new_intervention as ni
        iv = ni( camp )
    elif iv_name == "MaleCircumcision":
        from emodpy_hiv.interventions.malecirc import new_intervention as ni
        iv = ni( camp )
    elif iv_name == "HIVDrawBlood":
        from emodpy_hiv.interventions.drawblood import new_diagnostic as nd
        iv = nd( camp )
    elif iv_name == "HIVARTStagingByCD4Diagnostic":
        from emodpy_hiv.interventions.cd4diag import new_diagnostic as nd
        iv = nd( camp )
    elif iv_name == "HIVARTStagingCD4AgnosticDiagnostic":
        from emodpy_hiv.interventions.cd4agdiag import new_diagnostic as nd
        iv = nd( camp )
    elif iv_name == "HIVSigmoidByYearAndSexDiagnostic":
        from emodpy_hiv.interventions.sigmoiddiag import new_diagnostic as nd
        iv = nd( camp )
    elif iv_name == "HIVMuxer":
        print( "WARNING: IGNORING HIVMuxer" )
    else:
        if method_to_call is None:
            print( f"ERROR: No function found to create intervention {iv_name}." )
            return None
        try:
            iv = method_to_call( camp )
        except Exception as ex:
            print( f"Exception invoking method for iv_name {iv_name}." )
            print( str(ex) )
    return iv

def convert_delay( delay ):
    """
    Param comes in as, e.g., FIXED/30, and needs to be converted into, e.g., { "Distribution": "FIXED_DURATION", "Delayed_Period": 30 }
    """
    if delay is None:
        return None
    delay_dict_return = {}
    delay_parsed = delay.split( "/" )
    delay_dict_return[ "Delay_Distribution"] = f"{delay_parsed[0]}_DURATION"
    if delay.startswith( "FIXED" ):
        delay_dict_return["Delay_Period"]  = float(delay_parsed[1])
    return delay_dict_return 

def create_campaign_from_concise( ccdl ):
    # encode converts each ccdl line into a dictionary which we process here.
    mycamp = encode( ccdl )
    for enc_event in mycamp:
        # what
        iv_name = enc_event["iv_name"]
        payload = enc_event["payload"] if "payload" in enc_event else None
        if type(iv_name) is list:
            iv = list()
            for idx in range(len(iv_name)):
                act_iv_name = iv_name[idx]
                act_payload = payload[idx]
                intervention = iv_class_factory_ish( act_iv_name, act_payload )
                iv.append( intervention )
        else:
            iv = [iv_class_factory_ish( iv_name, payload )]

        # where
        nodes = enc_event["nodes"]

        # who
        coverage = enc_event["frac"]
        sex = enc_event["sex"] if "sex" in enc_event else "All"
        min_age = enc_event["minage"] if "minage" in enc_event else 0
        max_age = enc_event["maxage"] if "maxage" in enc_event else comm._MAX_AGE
        ips = enc_event["ips"] if "ips" in enc_event else None

        if "signal" in enc_event and enc_event["signal"]:
            signal = enc_event["signal"]
            # print( f"signal={signal}." )
            delay = enc_event["delay"] if "delay" in enc_event else None
            delay = convert_delay( delay )
            if delay:
                from emod_api.interventions.common import DelayedIntervention
                iv = [DelayedIntervention( camp, Configs=iv, Delay_Dict = delay )]
            event = comm.TriggeredCampaignEvent( camp, Start_Day=enc_event["start_day"], Node_Ids=nodes, Triggers=[signal], Event_Name="TCE", Intervention_List=iv, Demographic_Coverage=coverage, Property_Restrictions=ips, Target_Gender=sex, Target_Age_Min=min_age, Target_Age_Max=max_age )
            event.Event_Coordinator_Config.pop( "Targeting_Config" )
            try:
                camp.add( event )
            except Exception as ex:
                print( str( ex ) )
                print( event )
        else:
            if "reps" in enc_event:
                reps = int(enc_event[ "reps" ])
                gap = float(enc_event[ "gap" ])
            else:
                reps = 1
                gap = -1
            try:
                event = comm.ScheduledCampaignEvent( camp, Start_Day=enc_event["start_day"], Node_Ids=nodes, Intervention_List=iv, Number_Repetitions=reps, Timesteps_Between_Repetitions=gap, Demographic_Coverage=coverage, Target_Age_Min=min_age, Target_Age_Max=max_age, Property_Restrictions=ips, Target_Gender=sex )
            except Exception as ex:
                print( str( ex ) )
                import pdb
                pdb.set_trace()
            event.Event_Coordinator_Config.pop( "Targeting_Config" )
            camp.add( event )

    outfile = "test_campaign.json"
    print( f"\n\nWriting {outfile} to disk." )
    camp.save( outfile )
        

if __name__ == "__main__":
    if len(sys.argv)<2:
        print( "Usage: python -m emodpy_hiv.camp_creator -c /path/to/ccdl -s /path/to/schema.json" )
        sys.exit()
    schema_path = sys.argv[2]
    camp.set_schema( schema_path )
    camp.unsafe = True
    ccdl = sys.argv[1]
    create_campaign_from_concise( ccdl )

